import pathlib


def _is_path(obj: object) -> bool:
    """
    Return whether the given object looks like a path (str, pathlib.Path or pathlib2.Path)
    """
    # TODO
    # pytest uses pathlib2.Path objects on Python 3.5 for its tmp_path fixture.
    # On Python 3.6+, this function can be replaced with isinstance(obj, os.PathLike)
    import sys

    if "pathlib2" in sys.modules:
        import pathlib2  # type: ignore

        path_classes = [str, pathlib.Path, pathlib2.Path]
    else:
        path_classes = [str, pathlib.Path]
    return isinstance(obj, tuple(path_classes))


def shorten(s: str, n: int = 100) -> str:

    """Shorten string s to at most n characters, appending "..." if necessary."""
    if s is None:
        return None
    if len(s) > n:
        s = s[: n - 3] + "..."
    return s



from abc import ABC, abstractmethod
from typing import Iterator, Tuple

from dnaio import SequenceRecord


class SingleEndReader(ABC):
    @abstractmethod
    def __iter__(self) -> Iterator[SequenceRecord]:
        pass


class PairedEndReader(ABC):
    @abstractmethod
    def __iter__(self) -> Iterator[Tuple[SequenceRecord, SequenceRecord]]:
        pass


class SingleEndWriter(ABC):
    @abstractmethod
    def write(self, record: SequenceRecord) -> None:
        pass


class PairedEndWriter(ABC):
    @abstractmethod
    def write(self, record1: SequenceRecord, record2: SequenceRecord) -> None:
        pass



from os import PathLike
from typing import Union, BinaryIO, Optional

from xopen import xopen

from . import SequenceRecord
from ._util import _is_path
from .interfaces import SingleEndWriter


class FileWriter:
    """
    A mix-in that manages opening and closing and provides a context manager
    """

    def __init__(
        self,
        file: Union[PathLike, str, BinaryIO],
        *,
        opener=xopen,
        _close_file: Optional[bool] = None,
    ):
        if _is_path(file):
            self._file = opener(file, "wb")
            self._close_on_exit = True
        else:
            self._file = file
            self._close_on_exit = bool(_close_file)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{getattr(self._file, 'name', self._file)}')"

    def close(self) -> None:
        if self._close_on_exit:
            self._file.close()

    def __enter__(self):
        if self._file.closed:
            raise ValueError("I/O operation on closed file")
        return self

    def __exit__(self, *args):
        self.close()


class FastaWriter(FileWriter, SingleEndWriter):
    """
    Write FASTA-formatted sequences to a file.
    """

    def __init__(
        self,
        file: Union[PathLike, str, BinaryIO],
        *,
        line_length: Optional[int] = None,
        opener=xopen,
        _close_file: Optional[bool] = None,
    ):
        """

        Arguments:
            file: A path or an open file-like object
            line_length: Wrap sequence lines after this many characters (None disables wrapping)
            opener: If *file* is a path, this function is called to open it.
        """
        super().__init__(file, opener=opener, _close_file=_close_file)
        self.line_length = line_length if line_length != 0 else None

    def __repr__(self) -> str:
        return f"FastaWriter('{getattr(self._file, 'name', self._file)}')"

    def write(self, name_or_record, sequence: Optional[str] = None):
        """Write a record to the FASTA file.

        If only one parameter (name_or_record) is given, it must have
        attributes .name and .sequence, which are then used.
        Otherwise, the first parameter must be the name and the second
        the sequence.

        The effect is that you can write this:
        writer.write("name", "ACCAT")
        or
        writer.write(SequenceRecord("name", "ACCAT"))
        """
        if sequence is None:
            name = name_or_record.name
            sequence = name_or_record.sequence
        else:
            name = name_or_record

        if self.line_length is not None:
            self._file.write((">" + name + "\n").encode("ascii"))
            s = []
            for i in range(0, len(sequence), self.line_length):
                s.append(sequence[i : i + self.line_length] + "\n")
            self._file.write("".join(s).encode("ascii"))
        else:
            text = ">" + name + "\n" + sequence + "\n"
            self._file.write(text.encode("ascii"))


class FastqWriter(FileWriter, SingleEndWriter):
    """
    Write records in FASTQ format.

    FASTQ files are formatted like this::

        @read name
        AACCGGTT
        +
        FF,:F,,F
    """

    file_mode = "wb"

    def __init__(
        self,
        file: Union[PathLike, str, BinaryIO],
        *,
        two_headers: bool = False,
        opener=xopen,
        _close_file: Optional[bool] = None,
    ):
        """
        Arguments:
            file: A path or an open file-like object
            two_headers: If True, the header is repeated on the third line
                of each record after the "+".
            opener: If *file* is a path, this function is called to open it.
        """
        super().__init__(file, opener=opener, _close_file=_close_file)
        self._two_headers = two_headers
        # setattr avoids a complaint from Mypy
        setattr(
            self, "write", self._write_two_headers if self._two_headers else self._write
        )

    def __repr__(self) -> str:
        return f"FastqWriter('{getattr(self._file, 'name', self._file)}')"

    def write(self, record: SequenceRecord) -> None:
        """
        Dummy method to make it possible to instantiate this class.
        The correct write method is assigned in the constructor.
        """
        assert False

    def _write(self, record: SequenceRecord) -> None:
        """
        Write a record to the FASTQ file.
        """
        self._file.write(record.fastq_bytes())

    def _write_two_headers(self, record: SequenceRecord) -> None:
        """
        Write a record to the FASTQ file, repeating the header
        in the third line after the "+" .
        """
        self._file.write(record.fastq_bytes(two_headers=True))

    def writeseq(self, name: str, sequence: str, qualities: str) -> None:
        self._file.write(f"@{name:s}\n{sequence:s}\n+\n{qualities:s}\n".encode("ascii"))



"""
Sequence I/O: Read and write FASTA and FASTQ files efficiently
"""

__all__ = [
    "open",
    "SequenceRecord",
    "SingleEndReader",
    "PairedEndReader",
    "SingleEndWriter",
    "PairedEndWriter",
    "FastaReader",
    "FastaWriter",
    "FastqReader",
    "FastqWriter",
    "UnknownFileFormat",
    "FileFormatError",
    "FastaFormatError",
    "FastqFormatError",
    "InterleavedPairedEndReader",
    "InterleavedPairedEndWriter",
    "TwoFilePairedEndReader",
    "TwoFilePairedEndWriter",
    "read_chunks",
    "read_paired_chunks",
    "records_are_mates",
    "__version__",
]

from os import PathLike
from typing import Optional, Union, BinaryIO

from xopen import xopen

from ._core import (
    SequenceRecord,
)
from ._core import record_names_match  # noqa: F401  # deprecated
from ._core import records_are_mates
from .readers import FastaReader, FastqReader
from .writers import FastaWriter, FastqWriter
from .singleend import _open_single
from .pairedend import (
    TwoFilePairedEndReader,
    TwoFilePairedEndWriter,
    InterleavedPairedEndReader,
    InterleavedPairedEndWriter,
)
from .exceptions import (
    UnknownFileFormat,
    FileFormatError,
    FastaFormatError,
    FastqFormatError,
)
from .interfaces import (
    SingleEndReader,
    PairedEndReader,
    SingleEndWriter,
    PairedEndWriter,
)
from .chunks import read_chunks, read_paired_chunks
from ._version import version as __version__


# Backwards compatibility aliases
Sequence = SequenceRecord


def open(
    file1: Union[str, PathLike, BinaryIO],
    *,
    file2: Optional[Union[str, PathLike, BinaryIO]] = None,
    fileformat: Optional[str] = None,
    interleaved: bool = False,
    mode: str = "r",
    qualities: Optional[bool] = None,
    opener=xopen
) -> Union[SingleEndReader, PairedEndReader, SingleEndWriter, PairedEndWriter]:
    """
    Open sequence files in FASTA or FASTQ format for reading or writing.

    Parameters:

      file1:
        Path or an open file-like object. For reading single-end reads, this is
        the only required argument.

      file2:
        Path or an open file-like object. When reading paired-end reads from
        two files, set this to the second file.

      mode:
        Either ``'r'`` or ``'rb'`` for reading, ``'w'`` for writing
        or ``'a'`` for appending.

      interleaved:
        If True, then file1 contains interleaved paired-end data.
        file2 must be None in this case.

      fileformat:
        If *None*, the file format is autodetected from the file name
        extension. Set to ``'fasta'`` or ``'fastq'`` to not auto-detect.

      qualities:
        When mode is ``'w'`` and fileformat is *None*, this can be set
        to *True* or *False* to specify whether the written sequences will have
        quality values. This is is used in two ways:

        - If the output format cannot be determined (unrecognized extension
          etc), no exception is raised, but fasta or fastq format is chosen
          appropriately.

        - When False (no qualities available), an exception is raised when the
          auto-detected output format is FASTQ.
      opener: A function that is used to open file1 and file2 if they are not
        already open file-like objects. By default, ``xopen`` is used, which can
        also open compressed file formats.

    Return:
       A subclass of `SingleEndReader`, `PairedEndReader`, `SingleEndWriter` or
       `PairedEndWriter`.
    """
    if mode not in ("r", "rb", "w", "a"):
        raise ValueError("Mode must be 'r', 'rb', 'w' or 'a'")
    if interleaved and file2 is not None:
        raise ValueError("When interleaved is set, file2 must be None")

    if file2 is not None:
        if mode in "wa" and file1 == file2:
            raise ValueError("The paired-end output files are identical")
        if "r" in mode:
            return TwoFilePairedEndReader(
                file1, file2, fileformat=fileformat, opener=opener, mode=mode
            )
        append = mode == "a"
        return TwoFilePairedEndWriter(
            file1,
            file2,
            fileformat=fileformat,
            qualities=qualities,
            opener=opener,
            append=append,
        )
    if interleaved:
        if "r" in mode:
            return InterleavedPairedEndReader(
                file1, fileformat=fileformat, opener=opener, mode=mode
            )
        append = mode == "a"
        return InterleavedPairedEndWriter(
            file1,
            fileformat=fileformat,
            qualities=qualities,
            opener=opener,
            append=append,
        )

    # The multi-file options have been dealt with, delegate rest to the
    # single-file function.
    return _open_single(
        file1, opener=opener, fileformat=fileformat, mode=mode, qualities=qualities
    )
